(* AUS - the advanced update system
 * Copyright (C) 2004-2019  Heinz Deinhart
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *)

open Printf

open Util

(** aus main program
*)

(*** local types *************************************************************)

(** parsed from arguments and tells out what to do *)
type running_modes =
    RM_USAGE | RM_DISPLAY_VERSION | RM_UPGRADE_AND_INSTALL
  | RM_UPGRADE_NO_INSTALL | RM_INSTALL_NO_UPGRADE | RM_CONFIG_PKGSYS_ONLY
  | RM_SYNC_METADATA_ONLY | RM_FLUSH_CACHE_ONLY | RM_DELETE_KERNELS_ONLY
  | RM_INFO_MISSING | RM_INFO_LIST_GROUPS | RM_INFO_DUMP_GROUP
  | RM_INFO_KERNEL

(** used for location of problems in config files *)
type loc = Loc of int * string
type pkgname = string
type pkgarch = string

(** A resource can be a packages name or a group.
    For a package a certain architecture might be specified, if not
    than any architecture will match.
*)

type resource =
    Package of pkgname * pkgarch option * loc
  | Group of string * loc

(** a group is basically a list of resources with support for -= and +=
*)
type group = {
  mutable group_base: resource list; (** = list *)
  mutable group_base_loc: loc; (** location for problem resolving *)
  mutable group_plus: resource list; (** += list applied before -= *)
  mutable group_minus: resource list; (** -= list, applied after +=
					  accepts only packages no groups *)
  mutable group_description: string; (** description for humans *)
}

(** used by parser of config files/urls to store its results *)
type config = {
  mutable pkgsys: Pkgsys.packageSystem; (** pkgsys class for runtime binding *)
  mutable pkgsysname: string; (** name of the package system to use *)
  mutable pkgsys_cfg: string list; (** repository URL,
					 format is pkgsys dependent *)
  groups: (string, group) Hashtbl.t; (** hashtable to lookup all groups *)
  mutable packages: resource list; (** required resources tree *)
  mutable kmods: resource list; (** required kernel modules tree *)
  mutable ausconfig: string option; (** currently unused, maybe deprecated (fixme) *)
  mutable option_delete_old_kernels: bool; (** tells aus to delete old
					       kernels or not *)
  mutable option_keep_kernels: int; (** if old kernels should be deleted
					this tells aus how many old
					kernels to keep *)
  mutable option_keep_first_kernel: bool; (** tells aus always to keep the
					      first kernel as a backup *)
  mutable option_kernel_types: (string * string) list; (** list of kernel types
							   FIXME *)

  mutable option_run_at_boot: bool; (** tell aus (not) to run at boot time *)
  mutable option_run_at_shutdown: bool; (** tell aus (not) to run during
					    shutdown of system *)
  mutable option_admin_mail_addr: string option; (** mail address that will
						     get error messages *)
  mutable option_clean_cache: bool; (** tells aus (not) to clean package
					cache every time it runs *)
  mutable option_dont_use_ocarpml: bool; (** tells aus (not) to use ocarpml,
					     but run rpm manually *)
}

(*** constants ***************************************************************)

let whitespace_regexp = Str.regexp "[ \t]+"
let argv0 = Sys.argv.(0)

(*** local references ********************************************************)

(** the default location of the configuration file *)
let config_file = ref "/etc/aus.conf"

(** the default cache location of the remote file access module *)
let rcache_path = ref "/var/cache/aus"

(** this is automatically replaced by make dist *)
let my_version = "0.2.2"

(** stores the hostname of the current system *)
let my_hostname = ref "unknown"

(** a runtime instance of the rfile module
    FIXME: check if this should be launched later
*)
let rcache = ref (Rfile.make !rcache_path)

(* * list of all packages in system, used as a cache, too *)
(* fixme: this is unused for some (currently unknown) reason
let system_package_list: (string, string) Hashtbl.t option ref = ref None
*)
(** the configuration container *)
let config = {
  pkgsys = new Pkgsysdebug.pkgSysDebug;
  pkgsysname = "null";
  pkgsys_cfg = [];
  groups = Hashtbl.create 128;
  packages = []; kmods = []; ausconfig = None;
  option_delete_old_kernels = false;
  option_keep_first_kernel = false;
  option_keep_kernels = 1;
  option_kernel_types = [];
  option_run_at_boot = true;
  option_run_at_shutdown = false;
  option_admin_mail_addr = None;
  option_clean_cache = true;
  option_dont_use_ocarpml = false;
}

(*** exceptions **************************************************************)

(** raised when something really bad happens *)
exception Fatal_error of string

(** raised during parse of config files/urls *)
exception Parse_error of loc * string
  
(*** ocarpml init ************************************************************)

exception RPM_Error of (int * string)

let ocarpml_init () =
  failwith "ocarpml support tamporarly disabled"

(*
let ocarpml_init () =
  Rpm.read_config ();
  Rpm.set_error_callback (fun code string ->
    raise (RPM_Error (code, string)))
*)
(*** generic functions *******************************************************)

(** simple parser for resource strings
    everything with a @ at front is a group, everything else a package
*)
let resource_from_string loc str =
  if str.[0] = '@' then
    Group (Str.string_after str 1, loc)
  else
    try
      let name, arch = name_arch_from_string str
      in
	Package (name, arch, loc)
    with
	Invalid_argument x -> raise (Parse_error (loc, x))

(** convert a location to human readable string
*)
let loc_to_string (Loc (linenr, filename)) =
  filename ^ " at " ^ (string_of_int linenr)

(*** package calcluation and friends *****************************************)

(** expands all '/'s in pkgs with all kernel versions in kvers
    result will be a list of pkg names with
    length = len.(pkgs) * len.(kvers(,,x))
*)
let resource_expand_kmods pkgs (kvers : (string * string * string list) list) =
  let conv kver =
    Debug.print 3 ("resource_expand_kmods:conv: " ^ kver);
    let resource_to_expansion = function
	Package (pkg, arch, _) -> Str.global_replace (Str.regexp "%") kver pkg
      | Group (_, loc) -> raise
	  (Parse_error
	     (loc, "group references not allowed in kernel-modules"))
    in
      List.map resource_to_expansion pkgs
  in let kvers2 = (* result: string list *)
      List.flatten (List.map (fun (_,ka,kv) -> List.map (fun v -> v^ka) kv)
		      kvers)
  in
    List.flatten (List.map conv kvers2)

(** returns the running kernel name as reported by "uname -r"
*)
let get_system_running_kernel_version () =
  Debug.print 3 "starting get_system_running_kernel_version";
  try
    let ich = Unix.open_process_in "uname -r"
    in let result = input_line ich
    in
      ignore (Unix.close_process_in ich);
      result
  with
      _ -> raise (Fatal_error "get_system_running_kernel_version failed")

(** returns all running kernels in a tuple:
    kernel rpm name, kernel_postfix, versions
    example: "kernel-smp", "smp", ["2.6.1", "2.6.2"]
*)
let get_system_kernel_versions () =
  let get_versions (kname,kadder) =
    try
      let rec work ?(result = []) ch =
	try
	  match Str.split whitespace_regexp (input_line ch) with
	      [x] -> 	work ~result:(x :: result) ch
	    | ["package"; _; "is"; "not"; _] -> work ~result ch
	    | _ ->
		raise (Fatal_error "get_system_kernel_versions: parse error")
	with
	    End_of_file -> result
      in
      let ich =
	Unix.open_process_in
	  ("rpm -q "^kname^" --qf '%{version}-%{release}\n'")
      in let kvers = work ich
      in
	ignore (Unix.close_process_in ich);
	kname, kadder, kvers
    with
	Fatal_error x -> raise (Fatal_error x)
      | excpt ->
	  let msg = sprintf "exception in get_system_kernel_versions: %s"
	    (Printexc.to_string excpt);
	  in
	    raise (Fatal_error msg)
  in
    Debug.print 3 "starting get_system_kernel_versions";
    let result = List.map get_versions config.option_kernel_types
    and sprint_tuple (x, y, z) =
      (sprintf "%s/%s: " x y) ^ (String.concat " " z)
    in
      Debug.print 5 "found tuples:";
      if (List.length result) = 0 then
	Debug.print 5 "none found :(";
      if !Debug.debug_message_level > 9 then
	List.iter (fun t -> Debug.print 5 (sprint_tuple t)) result;
      result

(** invalidates the cached system package list
*)
(* TODO FIXME: appearantly unused
let invalidate_system_package_list () =
  system_package_list := None
*)

(** returns a list of installed packages or the cache if already filled
*)
let get_system_package_list_new () =
  failwith "ocarpml support temporarly disabled"

(*
let get_system_package_list_new () =
  Debug.print 2 "get_system_package_list_new called";
  let syspkgs = Hashtbl.create 256
  in let rec iter dbi =
    let fetchname hdr =
      match Rpm.header_get_entry hdr Rpmtag.NAME with
	| Some(_, Rpm.String(name)) -> name
	| _ -> raise (RPM_Error (0, "Rpm.header_get_entry/NAME failed"))
    and fetchversion hdr =
      match Rpm.header_get_entry hdr Rpmtag.VERSION with
	| Some(_, Rpm.String name) -> name
	| _ -> raise (RPM_Error (0, "Rpm.header_get_entry/VERSION failed"))
    and fetcharch hdr =
      match Rpm.header_get_entry hdr Rpmtag.ARCH with
	| Some(_, Rpm.String arch) -> Some arch
	| _ -> raise (RPM_Error (0, "Rpm.header_get_entry/ARCH failed"))
    and fetchprovides hdr =
      match Rpm.header_get_entry hdr Rpmtag.PROVIDENAME with
	| Some(_, Rpm.StringArray provides) -> provides
	| _ -> raise (RPM_Error (0, "Rpm.header_get_entry/PROVIDENAME failed"))
    in
      match Rpm.db_next_iterator dbi with
	| Some hdr ->
	    Hashtbl.add syspkgs (fetchname hdr) (fetcharch hdr);
	    Array.iter (fun p -> Hashtbl.add syspkgs p None)
	      (fetchprovides hdr);
	    iter dbi
	| None -> ()
  in
    match Rpm.db_open "" with
      | Some db ->
	  begin
	    match Rpm.db_init_iterator db Rpmtag.RPMDBI_PACKAGES "" with
	      | Some dbi -> iter dbi ; syspkgs
	      | None -> raise (RPM_Error (0, "Rpm.db_init_iterator failed"))
	  end
      | None -> raise (RPM_Error (0, "Rpm.db_open failed"))
*)

let get_system_package_list_old () =
  try
    Debug.print 2 "get_system_package_list_old called";
    (* let rpm_regexp = Str.regexp " = " *)
    let syspkgs = Hashtbl.create 256
    in let rec work ch =
      match
	try
	  Some (input_line ch)
	with
	  End_of_file -> None
      with
      | None -> ()
      | Some line ->
	  match Str.split whitespace_regexp line with
	    [name; "=)"; arch; "(="; version] ->
	      let arch = Some arch
	      in
	      Hashtbl.add syspkgs name arch; (* add with known arch *)
	      Debug.print 150 ("got pkg: " ^
			       (string_of_name_arch (name, arch)) ^
			       " [" ^ version ^ "]");
	      work ch
	  | name :: _ ->
	      Hashtbl.add syspkgs name None; (* add with unknown arch *)
	      Debug.print 150 ("got provides: " ^ name);
	      work ch
	  | _ ->
	      raise (Fatal_error
		       ("get_system_package_list: parse error: " ^ line))
    in
    Debug.print 2 "starting get_system_package_list";
    let ich = Unix.open_process_in
	"rpm -qa --provides --queryformat '%{name} =) %{arch} (= %{version}\n'"
    in
    work ich;
    ignore (Unix.close_process_in ich);
    syspkgs
  with
    Fatal_error x -> raise (Fatal_error x)
  | excpt ->
      let msg = sprintf "exception in get_system_package_list_old: %s"
	  (Printexc.to_string excpt);
      in
      raise (Fatal_error msg)

let get_system_package_list () =
  try
    if config.option_dont_use_ocarpml then
      get_system_package_list_old ()
    else
      let r = get_system_package_list_new ()
      in
	Gc.major ();
	r
  with
    | RPM_Error(_, msg) ->
	raise (Fatal_error
		  (sprintf "get_system_package_list_new failed: %s" msg))
    | excpt ->
	Debug.print 1
	  (sprintf "get_system_package_list_new failed (%s), falling back"
	      (Printexc.to_string excpt));
	get_system_package_list_old ()

(** Computes what packages are in a profile. Basically this fun flattens
    the profile tree by resolving all group contents and executing the
    += and -= statements
*)
let compute_packages config =
  let pkgs = Hashtbl.create 128
  and groups = Hashtbl.create 32
  in let rec compute_group pkglist g l =
      try
	ignore (Hashtbl.find groups g);
	pkglist
      with
	  Not_found ->
	    Hashtbl.add groups g l;
	    let pkglist2 =
	      add_resources [] (List.append g.group_base g.group_plus)
	    and minus_pkgs =
	      List.map (function
		| Package (p, a, _) -> p, a
		| _ -> raise (Fatal_error
				 "compute_group minus_pkgs should not happen"))
		g.group_minus
	    in let rec filter ?(result=[]) = function
	      | [] -> List.rev result
	      | p :: rest ->
		  if List.mem p minus_pkgs then
		    filter ~result rest
		  else
		    filter ~result:(p :: result) rest
	    in
		 List.append pkglist (filter pkglist2)
	    (* add_resources pkglist (List.append g.group_base g.group_plus) *)
     and add_resources pkglist = function
	 Package (p, a, l) :: rest ->
	   begin
	     try
	       ignore (Hashtbl.find pkgs (p, a));
	       add_resources pkglist rest
	     with
		 Not_found ->
		   Hashtbl.add pkgs (p, a) l;
		   add_resources ((p, a) :: pkglist) rest
	   end
       | Group (g, l) :: rest ->
	   begin
	     try
	       add_resources
		 (compute_group pkglist (Hashtbl.find config.groups g) l)
		 rest
	     with
		 Not_found ->
		   let msg = sprintf
		     "configuration error: unknown group '%s' used at %s"
		     g (loc_to_string l)
		   in
		     raise (Fatal_error msg)
	   end
       | [] ->
	   pkglist
  in
    Debug.print 2 "starting compute_packages";
    let result =
      List.fast_sort name_arch_compare (add_resources [] config.packages)
    in
      Debug.print 2 "finished compute_packages";
      result

(** Computes what packages are missing on a system. This is kind of tricky,
    because package names and architectures must be considered.
    If a profile requests a certain architecture then a package is flagged
    as missing if it is not installed with this very architecture. If a profile
    requests no architecture than any architecture is sufficient.
    @param sys_pkgs Hashtable with pkgnames as key and pkgarch option as value
    @param profile_pkgs List of pkgname * pkgarch option tuples of a profile
*)
let compute_missing (sys_pkgs : (pkgname, pkgarch option) Hashtbl.t) 
      profile_pkgs =
  let rec work result = function
    | (name, arch) :: rest -> (* any arch is sufficient *)
	let pkg = name, arch
	in begin
	  Debug.print 150 ("CM: now checking pkg " ^ (string_of_name_arch pkg)
			   ^ (sprintf "where name='%s' arch='%s'" name
				 (match arch with None -> "<<None>>" |
				     Some x -> x)));
	  try
	    ignore (Hashtbl.find sys_pkgs name); (* just for exception *)
	    let installed_archs : pkgarch option list = 
              Hashtbl.find_all sys_pkgs name
	    in
	      match arch with
		| None -> (* dont care bout arch -> already installed *)
		    Debug.print 150 "CM:   .. already installed for some arch";
		    work result rest
		| Some arc ->
		    if (List.mem arch installed_archs) then begin
		      Debug.print 150 "CM:   .. installed with exact arch";
		      work result rest (* already installed *)
		    end else begin
		      Debug.print 150 "CM:   .. not installed for that arch";
		      work (pkg :: result) rest (* this very arch is missing *)
		    end
	  with
	      Not_found ->
		Debug.print 150 ("CM:   .. not installed");
		work (pkg :: result) rest (* totally  missing *)
	end;
    | [] ->
	result
  in
    Debug.print 2 "starting compute_missing (CM)";
    let result =  List.fast_sort compare (work [] profile_pkgs)
    in
      Debug.print 2 (sprintf "finished compute_missing (%i pkgs)"
			(List.length result));
      result

(*** kernel cleanup tools ****************************************************)

let perform_kernel_cleanup ?(info_only=false) allversions =
  let running_version =
    get_system_running_kernel_version ()
  and all_postfixes = List.map (fun (_, x, _) -> x) allversions
  in let compute_obsolete_versions versions kpostfix =
      try
	let vsorted = List.sort Rpm.vercmp versions
	and keepfirst = config.option_keep_first_kernel
	and keepnr = config.option_keep_kernels
	in let rec vfirsted =
	    List.rev (if keepfirst then List.tl vsorted else vsorted)
	   and reduce found_running cnt ver_list =
	    Debug.print 5 (Printf.sprintf "reduce %B(%s) %i [%s]"
			 found_running running_version cnt
			 (String.concat "; " ver_list));
	    match found_running, ver_list, cnt with
		_, [], _ -> []
	      | false, v :: _, _ ->
		  if List.exists (fun x -> (v^x) = running_version)
		    all_postfixes then
		      reduce true cnt (List.tl ver_list)
		  else
		    reduce false cnt (List.tl ver_list)
	      | true, v, 0 -> ver_list
	      | true, v, _ -> reduce true (cnt - 1) (List.tl ver_list)
	in
	  reduce false keepnr vfirsted
      with
	  Failure _ -> []
     and expand_versions ((kname, kpostfix, kversions) as versions) =
      Debug.print 5 ("perform_kernel_cleanup:expand_versions now doing " ^
		   kname ^ "/" ^ kpostfix);
       List.append (List.map (fun kv -> kname ^ "-" ^ kv ) kversions)
	(resource_expand_kmods config.kmods [versions])
  in let do_version (kname, kpostfix, kvers) =
      let obsolete_versions = compute_obsolete_versions kvers kpostfix
      in
	Debug.print 5 ("perform_kernel_cleanup:do_version: " ^
			  kname ^ "/" ^ kpostfix);
	Debug.print 4 ("killing: " ^ (String.concat " " (expand_versions (kname, kpostfix, obsolete_versions))));
  in
    Debug.print 1 ("perform_kernel_cleanup: starting, running=" ^ running_version);
    List.iter do_version allversions

(*** configuration parsing ***************************************************)

let parse_pkgsys_config loc str =
  Debug.print 3 (sprintf "parse_pkgsys_config '%s'" str);
  config.pkgsys_cfg <- str :: config.pkgsys_cfg

(* hack to support obsolete repository config file option *)
let parse_repository loc str =
  Debug.print 3 (sprintf "parse_repository (OBSOLETE) '%s'" str);
  match Str.split whitespace_regexp str with
    | "rpm" :: baseurl :: adder :: comps ->
	let conv_repo comp =
	  parse_pkgsys_config loc
	    (sprintf "repository %s/%s/RPMS.%s" baseurl adder comp)
	in
	  List.iter conv_repo comps
    | _ -> raise (Parse_error (loc, sprintf "illegal 'repository' line"))

let parse_ausconfig loc str =
  Debug.print 3 (sprintf "parse_ausconfig '%s'" str);
  config.ausconfig <- Some str

let parse_packages loc str =
  Debug.print 3 (sprintf "parse_packages '%s'" str);
  config.packages <- List.map (resource_from_string loc)
    (Str.split whitespace_regexp str)

let parse_kmods loc str =
  Debug.print 3 (sprintf "parse_packages '%s'" str);
  config.kmods <- List.map (resource_from_string loc)
    (Str.split whitespace_regexp str)

let parse_option loc str =
  let parse_bool opname str =
    try
      bool_of_string (String.lowercase_ascii str)
    with
	_ -> raise (Parse_error (loc, sprintf
				   "option %s expects bool but got %s"
				   opname str))
  and parse_int ?(range=fun _ -> true) str opname =
    try
      let i = int_of_string str
      in
	if (range i) = false then
	  raise (Parse_error (loc, sprintf
				"options %s argument %s is out of range"
				opname str));
	i
    with
	_ -> raise (Parse_error (loc, sprintf
				   "option %s expects int but got %s"
				   opname str))
  in
    Debug.print 3 (sprintf "parsing option '%s'" str);
    match Str.split whitespace_regexp str with
      | ["package-system"; namestr] ->
	  config.pkgsysname <- namestr;
      | ["delete-old-kernels" as n; numstr] ->
	  config.option_delete_old_kernels <- parse_bool n numstr
      |  ["keep-kernels" as n; numstr] ->
	   config.option_keep_kernels <-
	     parse_int ~range:(fun x -> x >= 0) numstr n
      | ["keep-first-kernel" as n; numstr] ->
	  config.option_keep_first_kernel <- parse_bool n numstr
      | ["add-kernel-type"; kernel_name ] ->
	  config.option_kernel_types <-
	    (kernel_name, "") :: config.option_kernel_types
      | ["add-kernel-type"; kernel_name; uname_postfix ] ->
	  config.option_kernel_types <-
	  (kernel_name, uname_postfix) :: config.option_kernel_types
      | ["admin-mail"; mail_addr] ->
	  config.option_admin_mail_addr <- Some mail_addr
      | ["run-at-boot" as n; boolstr] ->
	  config.option_run_at_boot <- parse_bool n boolstr
      | ["run-at-shutdown" as n; boolstr] ->
	  config.option_run_at_boot <- parse_bool n boolstr
      | ["clean-cache" as n; boolstr] ->
	  config.option_clean_cache <- parse_bool n boolstr
      | ["dont-use-ocarpml" as n; boolstr] ->
	  config.option_dont_use_ocarpml <- parse_bool n boolstr
      | _ -> raise (Parse_error (loc, "illegal option: "^str))

let get_group name =
  try
    Hashtbl.find config.groups name
  with
      Not_found ->
	let g = { group_base = []; group_base_loc = (Loc (0, "unknown"));
		  group_plus = []; group_minus = []; group_description = ""; }
	in
	  Hashtbl.add config.groups name g;
	  g

let parse_groupdesc loc str =
  Debug.print 4 (sprintf "parse_groupdesc '%s'" str);
  match Str.bounded_split whitespace_regexp str 3 with
      [name; "="; desc] ->
	begin
	  try
	    let g = Hashtbl.find config.groups name
	    in
	      g.group_description <- desc
	  with
	      Not_found ->
		raise (Parse_error
			 (loc, sprintf
			    "group description for unknown group %s" name))
	end
    | _ -> raise (Parse_error (loc, "illegal groupdesc usage"))

let parse_group loc str =
  Debug.print 4 (sprintf "parse_group '%s'" str);
  match Str.split whitespace_regexp str with
      name :: "=" :: rest ->
	let g = get_group name
	in
	  if g.group_base = [] then
	    begin
	      g.group_base <- List.map (resource_from_string loc) rest;
	      g.group_base_loc <- loc
	    end else
	      raise (Parse_error
		       (loc, sprintf
			  "duplicate group %s previously defined at %s"
			  name (loc_to_string g.group_base_loc)))
    | name :: "+=" :: rest ->
	let g = get_group name
	in
	  g.group_plus <- List.append g.group_plus
	    (List.map (resource_from_string loc) rest)
    | name :: "-=" :: rest ->
	let g = get_group name
	in let package_resource_from_string str =
	  match resource_from_string loc str with
	    | Group _ ->
		raise (Parse_error (loc, "groups are not allowed within -="))
	    | p -> p
	in
	     g.group_minus <- List.append g.group_minus
	       (List.map package_resource_from_string rest)
    | _ :: x :: _ ->
	raise (Parse_error (loc, sprintf
			      "expecting [+-]?= instead of %s" x))
    | _ ->
	raise (Parse_error (loc, "illegal group definition"))

let parse_at loc str =
  Debug.print 3 (sprintf "parse_at '%s'" str);
  match Str.bounded_split whitespace_regexp str 2 with
    host :: rest :: [] ->
      if host = !my_hostname then begin
	parse_group loc rest;
	Debug.print 5 ( sprintf "parsing group for host %s" host)
      end else
	Debug.print 5 ( sprintf "skipping group for host %s" host)
  | _ ->
      raise (Parse_error (loc, "illegal at definition"))

let rec parse_include ?(ignore_failure=false) loc str =
  let split_regexp = Str.regexp "[ \t]+"
  in let rec do_include = function
      url :: rest ->
	begin
	  try
	    parse_url ~filename:url (Rfile.open_in_channel !rcache url)
	  with
	      Rfile.Url_not_found url ->
		do_include rest
	end
    | [] ->
	if ignore_failure == true then ()
        else raise (Parse_error (loc, "include: no file found\n"))
  in
    do_include (Str.split split_regexp str)
and parse_url ?(linenr = 1) ~filename ich =
  let replace_specials str =
    let host_regexp = Str.regexp "%[Hh]"
    and prozent_regexp = Str.regexp "%%"
    and comment_regexp = Str.regexp "^#.*$"
    in let str = Str.global_replace host_regexp !my_hostname str
    in let str = Str.global_replace prozent_regexp "%" str
    in
      Str.global_replace comment_regexp "" str
(*
  and line_regexp = Str.regexp "^[ \\t]*\\([^ \\t]+\\)[ \\t]+\\(.+\\)$"
  and nop_regexp = Str.regexp "^[ \\t]*$"
*)
  in
    try
      let xline = input_line ich in
	Debug.print 10 ("parse_url: input_line returned '" ^ xline ^ "'");
      let line = replace_specials xline in
	Debug.print 150 (sprintf "now matching '%s'.." line);
	begin
	  match Str.bounded_split whitespace_regexp line 2 with
	      ["include"; data] ->
		parse_include (Loc (linenr, filename)) data
	    | ["!include"; data] ->
		parse_include ~ignore_failure:true
		  (Loc (linenr, filename)) data
	    | ["option"; data] ->
		parse_option (Loc (linenr, filename)) data
 	    | ["ausconfig"; data] ->
		parse_ausconfig (Loc (linenr, filename)) data
	    | ["packages"; data] ->
		parse_packages (Loc (linenr, filename)) data
	    | ["kernel-modules"; data] ->
		parse_kmods (Loc (linenr, filename)) data
	    | ["pkgsys"; namestr] ->
		config.pkgsysname <- namestr;
	    | ["pkgsys-config"; rest] ->
		parse_pkgsys_config (Loc (linenr, filename)) rest
	    | ["repository"; data] ->
		(* obsolete, backward compat for aus < v0.2 *)
		parse_repository (Loc (linenr, filename)) data
	    | ["group"; data] -> parse_group (Loc (linenr, filename)) data
	    | ["group-desc"; data] ->
		parse_groupdesc (Loc (linenr, filename)) data
	    | ["at"; data] -> parse_at (Loc (linenr, filename)) data
	    | [] -> ()
	    | [x; _] ->
		raise (Parse_error (Loc (linenr, filename),
				    sprintf "unknown keyword: %s" x))
	    | _ -> raise (Parse_error (Loc (linenr, filename), "syntax error"))
	end;
	parse_url ~linenr:(linenr+1) ~filename ich
    with
	End_of_file -> ()

let read_config config_file =
  try
    parse_url ~filename:config_file
      (Rfile.open_in_channel !rcache config_file)
  with
      Rfile.Url_not_found config_file ->
	raise (Fatal_error (sprintf "config file '%s' not found" config_file))
    | Rfile.Illegal_url url ->
	raise (Fatal_error (sprintf "illegal url '%s'" url))
    | Rfile.Connection_failure url ->
	raise (Fatal_error (sprintf "connection failure to '%s'" url))
    | Rfile.Cache_failure str ->
	raise (Fatal_error (sprintf "cache failure '%s'" str))

let read_repo_config url =
  try
    parse_url ~filename:url (Rfile.open_in_channel !rcache url);
    Debug.print 5 (Printf.sprintf "%s: included %s\n" argv0 url)
  with
    | Rfile.Url_not_found url ->
	Debug.print 3 (Printf.sprintf "%s: could not find %s\n"
			  argv0 url)
    | Rfile.Connection_failure url ->
	Debug.print 3 (Printf.sprintf "%s: could not find %s\n"
			  argv0 url)


(*** run_ funs: they perform the job *****************************************)

(** The (created) fun is called back whenever some progress is to be
    announced to the user.
*)
let create_progress_callback () =
  let last_string = ref ""
  and vindex = ref (-1)
  in let callback pos msg =
    let pos = if pos < 0.0 then 0.0 else if pos > 1.0 then 1.0 else pos
    in let str = sprintf "%i%% (%s)" (int_of_float (pos *. 100.0)) msg;
    in
      if str != !last_string then begin (* only update if info changed *)
	if !vindex != (-1) then
	  Debug.vsmalldel !vindex;
	vindex := Debug.vsmallmsg str;
	last_string := str;
      end
  in
    callback

let run_upgrade () =
  let vx = Debug.vsmallmsg "upgrading system.. "
  in
    Debug.vprint "upgrading system..";
    config.pkgsys#upgrade ~progress_callback:(create_progress_callback ()) ();
    Debug.vsmalldel vx

and run_install profile_pkgs =
  let system_pkgs = begin
    Debug.vprint "scanning installed packages";
    let vx = Debug.vsmallmsg "scanning installation..";
    in let retval = get_system_package_list ()
    in
      Debug.vsmalldel vx;
      retval
    end
  and kvers = get_system_kernel_versions ()
  in let vx = Debug.vprint "installing missing packages..";
       Debug.vsmallmsg "installing missing packages.. "
  in let missing_pkgs = compute_missing system_pkgs
    (List.append profile_pkgs [])
(*	(resource_expand_kmods config.kmods kvers)) FIXME *)
  in
    if (List.length missing_pkgs) > 0 then
      config.pkgsys#install ~progress_callback:(create_progress_callback ())
	missing_pkgs;
    Debug.vsmalldel vx

and run_activate_configuration () =
  let vx = Debug.vsmallmsg "activating pkgsys config.."
  in
    Debug.vprint "activating pkgsys config";
    config.pkgsys#activate_configuration ();
    Debug.vsmalldel vx

and run_sync_metadata () =
  let vx = Debug.vsmallmsg "synchronizing metadata.. "
  in
    Debug.vprint "synchronizing metadata";
    config.pkgsys#update_metadata
      ~progress_callback:(create_progress_callback ()) ();
    Debug.vsmalldel vx

and run_flush_cache enforce =
  if enforce || config.option_clean_cache then
    let vx = Debug.vsmallmsg "cleaning cache"
    in
      Debug.vprint "cleaning cache";
      config.pkgsys#clean ();
      Debug.vsmalldel vx
  else
    Debug.print 2 "skipping clean cache because of config option"

and run_dist_clean () =
  let vx = Debug.vsmallmsg "distcleaning cache"
  in
    Debug.vprint "distcleaning cache";
    config.pkgsys#distclean ();
    Debug.vsmalldel vx

and run_delete_kernels ?(info_only=false) () =
  let vx = Debug.vsmallmsg "deleting old kernels"
  in
    Debug.vprint "deleting old kernels";
    perform_kernel_cleanup ~info_only (get_system_kernel_versions ());
    Debug.vsmalldel vx

and run_info_missing profile_pkgs =
  let system_pkgs = get_system_package_list ()
  and kvers = get_system_kernel_versions ()
  and print_info config profile_pkgs system_pkgs missing_pkgs =
    Debug.vprint ~lvl:1 "Packages in profile: ";
    List.iter (fun na -> Debug.vprint ~lvl:1 (" " ^ (string_of_name_arch na)))
      profile_pkgs;
    Debug.vprint ~lvl:1 "\n";
    Debug.vprint ~lvl:1 "\nPackages missing on system:\n";
    List.iter (fun na -> print_string ((string_of_name_arch na)^"\n"))
      missing_pkgs;
    flush stdout (* this should be outputted always *)
  in let missing_pkgs = compute_missing system_pkgs
    (List.append profile_pkgs [])
    (*	 (resource_expand_kmods config.kmods kvers)) *)
  in
    print_info config profile_pkgs system_pkgs missing_pkgs

and run_info_list_groups () =
  Debug.vprint ~lvl:1 "Annotated List of groups:\n";
  let print_fun = if !Debug.verbose_level > 0 then
    (fun (n, g) -> printf "%s: %s\n" n g.group_description)
    else
      (fun (n, _) -> printf "%s\n" n)
  in let groups =Hashtbl.fold (fun a b l -> (a,b) :: l) config.groups []
  in let group_cmp (n1,_) (n2,_) = compare_case_insensitive n1 n2
  in
       List.iter print_fun (List.fast_sort group_cmp groups)

(* FIXME: does not handle -= correct *)
and run_info_dump_group groupname =
  try
    let g = Hashtbl.find config.groups groupname
    and print_resource_pkg = function
	Package (n, a, _) -> print_string (" " ^ (string_of_name_arch (n,a)))
      | Group _ -> ()
    in
      Debug.vprint ~lvl:1 (groupname ^ ":");
      List.iter print_resource_pkg (List.append g.group_base g.group_plus);
      print_string "\n";
  with
      Not_found ->
	raise (Fatal_error (sprintf
			       "group %s not found in config\n"
			       groupname))

(*** "main" program **********************************************************)

let clean_exit () =
  Debug.vsmalldel 0;
  exit 0

let _ =
  try
    let running_mode = ref RM_USAGE
    and at_boot = ref false
    and at_shutdown = ref false
    (* and check_sys_msg = "checking system.." *)
    and group_to_dump = ref "@internal error"
    and pkgsys_override = ref None
    in let set_running_mode m () =
	if (!running_mode != RM_USAGE) then
	  begin
	    Debug.print 1 "set_running_mode failed";
	    raise (Arg.Bad "Please use only one of -a -u -n -c -k -s");
	  end;
	 running_mode := m
    in let speclist = [
      ("--config-file", Arg.Set_string config_file,
       (sprintf "configuration file (default: %s)" !config_file));
      ("--cache-path", Arg.Set_string rcache_path,
       (sprintf "cache path (default: %s)" !rcache_path));
      ("-d", Arg.Set_int Debug.debug_message_level, "debug level");
      ("-v", Arg.Unit (fun () -> Debug.verbose_level :=
	1 + !Debug.verbose_level),
      "increase verbosity");
      ("-q", Arg.Unit (fun () -> Debug.verbose_level :=
	!Debug.verbose_level - 1),
      "decrease verbosity");
      ("-i", Arg.String
	(function
	    "missing" | "m" -> set_running_mode RM_INFO_MISSING ()
	  | "kernel" | "k" -> set_running_mode RM_INFO_KERNEL ()
	  | "groups" | "g" -> set_running_mode RM_INFO_LIST_GROUPS ()
	  | grp when grp.[0] == '@' ->
	      set_running_mode RM_INFO_DUMP_GROUP ();
	      group_to_dump := String.sub grp 1 ((String.length grp)-1)
	  | _ -> raise (Arg.Bad
			   "-I wants: missing|kernel|groups|@<groupname>")),
      "get info, args: missing|kernel|groups|@<groupname>");
      ("-P", Arg.String (fun x -> pkgsys_override := Some x),
       "sets name of the package system and overrides config file setting");
      ("-R", Arg.Unit (fun () -> config.option_dont_use_ocarpml <- true),
       "do not use the ocarpml, instead call rpm manually");
      ("-a", Arg.Unit (set_running_mode RM_UPGRADE_AND_INSTALL),
       "perform full upgrade and install run (implies -c -k -s)");
      ("-u", Arg.Unit (set_running_mode RM_UPGRADE_NO_INSTALL),
       "like -a but does not install missing packages");
      ("-n", Arg.Unit (set_running_mode RM_INSTALL_NO_UPGRADE),
       "like -A but skips the generic upgrade before install");
      ("-c", Arg.Unit (set_running_mode RM_FLUSH_CACHE_ONLY),
       "just clean package cache, even if disabled in config");
      ("-k", Arg.Unit (set_running_mode RM_DELETE_KERNELS_ONLY),
       "only delete old kernels");
      ("-s", Arg.Unit (set_running_mode RM_SYNC_METADATA_ONLY),
       "only synchronize metadata from repositories (implies -l)");
      ("-l", Arg.Unit (set_running_mode RM_CONFIG_PKGSYS_ONLY),
       "only configure the local package management system");
      ("--at-boot", Arg.Unit (fun () -> at_boot := true),
       "tells aus that it is running during system boot");
      ("--at-shutdown", Arg.Unit (fun () -> at_shutdown := true),
       "tells aus that it is running during system shutdown");
      ("--version", Arg.Unit (set_running_mode RM_DISPLAY_VERSION),
       "print version information");
    ]
    and usage_message =
      argv0 ^ ": " ^ argv0 ^ " -[auncks] [ {option} [..] ]\n"
    in
      Arg.parse speclist
	(fun s -> Arg.usage speclist (sprintf "illegal argument: %s" s))
	usage_message;
      Debug.print 1 "arguments parsed";
      if !running_mode = RM_DISPLAY_VERSION then
	begin (* in that case abort early *)
	  fprintf stderr "%s: version %s\n" argv0 my_version;
	  clean_exit ()
	end;
      Debug.print 2 "getting hostname";
      my_hostname := get_my_hostname ();
      Debug.print 2 "calling Rfile.make";
      rcache := Rfile.make !rcache_path;
      Debug.print 1 "calling read_config";
      read_config !config_file;
      Debug.print 2 "checking for --at-*";
      if !at_boot then
	if config.option_run_at_boot == false then clean_exit ();
      if !at_shutdown then
	if config.option_run_at_shutdown == false then clean_exit ();
      Debug.print 2 "initializing package system";
      begin
	let psname = match !pkgsys_override, config.pkgsysname with
	  | Some pn, _ -> pn
	  | None, pn -> pn
	in
	  match psname with
	    | "debug" -> config.pkgsys <- new Pkgsysdebug.pkgSysDebug
	    | "apt" -> config.pkgsys <- new Pkgsysapt.pkgSysApt
	    | "yum" -> config.pkgsys <- new Pkgsysyum.pkgSysYum
	    | "smart" -> config.pkgsys <- new Pkgsyssmart.pkgSysSmart
	    | "null" ->
		raise (Fatal_error "specifying a package system is mandatory")
	    | _ ->
		raise (Fatal_error
			  (sprintf "failed to setup package system %s" psname))
      end;
      Debug.print 3 "configuring package system";
      config.pkgsys#configure (List.rev config.pkgsys_cfg);
      (* now we read additional (optional) include files that are
	 located in the accessed repositories

	 FIXME TODO: the problem here is that with backends we know
	 nothing about that anymore ... hmm
      *)
      Debug.print 1 "reading repository configs";
      List.iter read_repo_config (config.pkgsys#additional_includes ());
      Debug.print 1 "computing repository packages";
      let profile_pkgs = compute_packages config
      in
	Debug.print 1 "starting to run";
	begin
	  match !running_mode with
	      RM_DELETE_KERNELS_ONLY ->
		run_delete_kernels ();
		clean_exit ()
	    | RM_INFO_KERNEL ->
		run_delete_kernels ~info_only:true ();
		clean_exit ()
	    | RM_FLUSH_CACHE_ONLY ->
		run_flush_cache true;
		clean_exit ()
	    | RM_USAGE ->
		fprintf stderr "%s" usage_message;
		fprintf stderr "%s: to get help use: --help\n" argv0;
		exit 1
	    | RM_INFO_LIST_GROUPS ->
		run_info_list_groups ();
		clean_exit ()
	    | RM_INFO_MISSING ->
		run_info_missing profile_pkgs;
		clean_exit ()
	    | RM_INFO_DUMP_GROUP ->
		run_info_dump_group !group_to_dump;
		clean_exit ()
	    | _ -> ()
	end;
	Debug.print 2 "activating package system configuration";
	run_activate_configuration ();
	if !running_mode = RM_CONFIG_PKGSYS_ONLY then
	  clean_exit ();
	Debug.print 2 "run_sync_metadata";
	run_sync_metadata ();
	if !running_mode = RM_SYNC_METADATA_ONLY then
	  clean_exit ();
	if not config.option_dont_use_ocarpml then
	  begin
	    Debug.print 2 "initializing ocarpml library";
	    ocarpml_init ();
	  end;
	if (!running_mode = RM_UPGRADE_AND_INSTALL) ||
	  (!running_mode = RM_UPGRADE_NO_INSTALL) then
	  begin
	    run_upgrade ();
	    run_flush_cache false;
	  end;
	if (!running_mode = RM_INSTALL_NO_UPGRADE) ||
	  (!running_mode = RM_UPGRADE_AND_INSTALL) then
	  begin
	    run_install profile_pkgs;
	    run_flush_cache false;
	  end;
	run_delete_kernels ();
	Debug.vsmalldel 0; (* clear all output *)
	clean_exit ()
  with
   | Parse_error (Loc (line_nr, filename), msg) ->
       Debug.vsmalldel 0; (* clean up stdout *)
       fprintf stderr
	 "%s: parse error in config file %s at line %i: %s\n"
	 argv0 filename line_nr msg;
       exit 3
   | Fatal_error msg ->
       Debug.vsmalldel 0; (* clean up stdout *)
       fprintf stderr "%s: fatal error: %s\n" argv0 msg;
       exit 4
   | x ->
       Debug.vsmalldel 0; (* clean up stdout *)
       fprintf stderr "%s: FATAL INTERNAL ERROR: unhandled exception: %s\n"
	 argv0 (Printexc.to_string x);
       exit 9
