(* $Id: pkgsysapt.ml,v 1.9 2016/01/08 14:25:38 heinz Exp $
 *
 * AUS - the advanced update system
 * Copyright (C) 2004-2005  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Pkgsys

type repository = {
  repo_uri: string;
  repo_dist: string;
  repo_comps: string list;
}

let whitespace_regexp = Str.regexp "[ \t]+"

let url_separator_1 = Str.regexp "[/]"
let url_separator_2 = Str.regexp "[/]RPMS[.]"

(** Creates a baseuri distri comps style repository specification
    out of a real url that must contain "/RPMS.".
    This is an APT specific hack.
*)
let repo_of_url url =
  match Str.bounded_split url_separator_1 url 4 with
    | protocol :: "" :: host :: path :: [] -> begin
	match Str.bounded_split url_separator_2 path 2 with
	  | dist :: comp :: [] ->
	      { repo_uri = protocol ^ "//" ^ host;
		repo_dist = dist;
		repo_comps = [ comp ]
	      }
	  | _ -> raise (Configuration_error "repo_of_url failed")
      end
    | _ -> raise (Configuration_error "repo_of_url failed")

(** Converts "real urls" (see repo_of_url) into a list of
    baseuri distri comps style repository specifications.
    Same baseuri/distri combinations are grouped with multiple componentes.
    This is an APT specific hack.
*)
let repos_of_urls urls =
  let repos = Hashtbl.create 8
  and result = ref []
  in let add_url url =
    if Hashtbl.mem repos (url.repo_uri, url.repo_dist) then
      begin
	let url2 = Hashtbl.find repos (url.repo_uri, url.repo_dist)
	in
	  Hashtbl.replace repos (url.repo_uri, url.repo_dist)
	    { url with
	      repo_comps = List.append url.repo_comps url2.repo_comps }
      end
    else
      Hashtbl.add repos (url.repo_uri, url.repo_dist) url
  in
    List.iter (fun x -> add_url (repo_of_url x)) urls;
    Hashtbl.iter (fun _ x -> result := x :: !result) repos;
    !result

(** callback parser for apt output to create status messages *)
let create_trace_callback progress_callback =
  let current_job = ref (-1)
  and total_jobs = ref (-1)
  and already_installing = ref false
  and msg = ref "initialzing"
  and rx_predown = Str.regexp "^[0-9]+ upgraded, [0-9]+ newly installed"
  and rx_down = Str.regexp "^Get:[0-9]+"
  and rx_preinst = Str.regexp "^Committing changes..."
  and rx_inst = Str.regexp "^Preparing packages for installation"
  in let extract_nrs str =
    try
      begin
	match Str.split whitespace_regexp str with
	    ups :: "upgraded," :: ins :: _ ->
	      total_jobs := (int_of_string ups) + (int_of_string ins);
	      if !total_jobs < 1 then
		total_jobs := 1
	  | _ -> ()
      end
    with
	_ -> ()
  in let renew_pos () =
    let pos =
      if (!current_job >= 0) && (!total_jobs > 0) then
	(float_of_int !current_job) /. (float_of_int !total_jobs)
      else
	(0.0)
    in
      progress_callback pos !msg
  in let callback str =
    if (Str.string_match rx_predown str 0) then
      begin
	(* before download *)
	extract_nrs str;
	msg := "downloading";
	current_job := 0;
	renew_pos ()
      end
    else if Str.string_match rx_preinst str 0 then
      begin
	(* start of install *)
	msg := "preparing to install..";
	current_job := 0;
	renew_pos ()
      end
    else if Str.string_match rx_inst str 0 then
      begin
	msg := "installing";
	already_installing := true;
      end
    else if Str.string_match rx_down str 0 || !already_installing then
      begin
	incr current_job;
	renew_pos ()
      end
    else Debug.print 1 ("unparsed line: " ^ str)
  in
       callback

class pkgSysApt = object(this)
  inherit Pkgsys.packageSystem

  val mutable repo_urls = []

  method init () = repo_urls <- []

  method additional_includes () =
    List.rev (List.map (fun x -> x ^ "/ausgroups.conf") repo_urls)

  method configure msgs =
    let cfg_msg msg =
      match Str.split whitespace_regexp msg with
	| "repository" :: url :: [] -> repo_urls <- url :: repo_urls
	| _ -> ()
    in
      List.iter cfg_msg msgs

  method activate_configuration () =
    let repositories = repos_of_urls repo_urls
    in let och = open_out "/etc/apt/sources.list"
    in let rec work = function
	r :: rest -> output_string och
	  ("rpm " ^ (String.concat " "
			(r.repo_uri :: r.repo_dist :: r.repo_comps) ^ "\n"));
	  work rest
      | [] -> ()
    in
      work repositories;
      close_out och

  method update_metadata ?(progress_callback=(fun _ _ -> ())) () =
    try
      let retval, output =
	Job.launch "apt-get" [|"aus:apt-get"; "-y"; "update"|] (fun _ -> ())
      in
      if retval != 0 then
	let msg =
	  Printf.sprintf "apt/update_metadata: apt-get -y update failed: %i"
	    retval
	in
	raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "apt/update_metadata: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "apt/update_metadata: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method upgrade ?(progress_callback=(fun _ _ -> ())) () =
    try
      Debug.print 4 "pkgsysapt: upgrade called";
      let retval, output =
	Job.launch "apt-get" [|"aus:apt-get"; "-fmy";  "dist-upgrade"|]
	  (create_trace_callback progress_callback)
      in
      if retval != 0 then
	let msg =
	  Printf.sprintf "apt/upgrade: apt-get -fmy dist-upgrade failed: %i"
	    retval
	in
	raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "apt/upgrade: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "apt/upgrade: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method install ?(progress_callback=(fun _ _ -> ())) pkg_list =
    try
      Debug.print 4 "pkgsysapt: install called";
      let remove_arch = function
	| name, None -> name
	| n, Some _ ->
	    raise (Runtime_error ("apt/install: profiles with architectures" ^
				     " are not supported by apt (pkg " ^
				     n ^ ")", []))
      in let retval, output =
	Job.launch "apt-get"
	  (Array.of_list ("aus:apt-get" :: "-fmy" ::  "install" ::
			     (List.map remove_arch pkg_list)))
	  (create_trace_callback progress_callback)
      in
      if retval != 0 then
	let msg =
	  Printf.sprintf "apt/install: apt-get -fmy install failed: %i" retval
	in
	raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "apt/install: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "apt/install: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method remove _ (* pkg_list *) =
    raise (Runtime_error ("apt/remove: Not yet implemented", []))

  method clean () =
    try
      let retval, output =
	Job.launch "apt-get" [|"aus:apt-get"; "clean"|] (fun _ -> ())
      in
	if retval != 0 then
	  let msg = Printf.sprintf "apt/clean: apt-get clean failed: %i" retval
	  in
	  raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "apt/clean: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "apt/clean: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method distclean () =
    this#clean ()

end
