(* $Id: pkgsyssmart.ml,v 1.2 2016/01/08 14:25:38 heinz Exp $
 *
 * AUS - the advanced update system
 * Copyright (C) 2004-2005  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Printf

open Pkgsys
open Util

(* neede previous to ocaml 3.10 *)
let is_directory str =
  (Unix.stat str).Unix.st_kind = Unix.S_DIR

let whitespace_regexp = Str.regexp "[ \t]+"

let string_of_strings strs =
  List.fold_right (fun astr strs -> astr ^ "," ^ strs) strs ""

(** FIXME make this a config option *)
let smart_config_location = "/etc/smart/channels"

let trace_update_callback progress_callback =
  let fetching = ref false
  and rx_cache = Str.regexp "^Loading cache"
  and rx_fetch = Str.regexp "^Fetching information"
  and rx_status = Str.regexp "^.*[[] *\\([0-9]+\\)[%]]$"
  and extract_number line = (* string_match must be called before! *)
    (float_of_string (Str.matched_group 1 line)) /. 100.0
  in let closure line =
    if Str.string_match rx_cache line 0 then
      progress_callback 0.0 "updating cache"
    else if Str.string_match rx_fetch line 0 then
      fetching := true;
    if !fetching && Str.string_match rx_status line 0 then
      progress_callback (extract_number line) "downloading"
  in
       closure

let trace_transaction_callback progress_callback =
  let halfadder = ref 0.0
  in let fetching = ref false
  and msg = ref "downloading"
  and rx_cache = Str.regexp "^Loading cache"
  and rx_computing = Str.regexp "^Computing transaction"
  and rx_fetch = Str.regexp "^(Upgrad|Install)ing packages"
  and rx_install = Str.regexp "^Committing transaction"
  and rx_status = Str.regexp "^.*[[] *\\([0-9]+\\)[%]]$"
  and extract_number line = (* string_match must be called before! *)
    (float_of_string (Str.matched_group 1 line)) /. 200.0 +. !halfadder
  in let closure line =
    Debug.print 99 ("SMART-INST: " ^ line);
    if Str.string_match rx_cache line 0 then
      progress_callback 0.0 "updating cache"
    else if Str.string_match rx_computing line 0 then
      progress_callback 0.0 "computing transaction"
    else if Str.string_match rx_fetch line 0 then
      fetching := true
    else if Str.string_match rx_install line 0 then begin
      fetching := true; (* useful if there was no download *)
      halfadder := 0.5;
      msg := "installing"
    end;
    if !fetching && Str.string_match rx_status line 0 then
      progress_callback (extract_number line) !msg;
  in
       closure

class pkgSysSmart = object
  inherit packageSystem

  val mutable repo_urls = []

  method init () = repo_urls <- []

  method additional_includes () =
    List.rev (List.map (fun x -> x ^ "/ausgroups.conf") repo_urls)

  method configure msgs =
    let cfg_msg msg =
      match Str.split whitespace_regexp msg with
	| "repository" :: url :: [] -> repo_urls <- url :: repo_urls
	| _ -> ()
    in
      List.iter cfg_msg msgs

(*
  okay, bei smart laeufts etwas anders..
  das austauschen von channels laeuft so ab:

    smart channel --remove-all --yes
    smart channel --add aus.channel --yes
    smart channel --yes --add name type=rpm-md baseurl=http://

*)

  method activate_configuration () =
    let delete_configuration () =
      Debug.print 4 "pkgsyssmart: delete_configuration called";
      try
	let retval, output =
	  Job.launch "smart" [|"aus:smart"; "channel"; "--remove-all";
			       "--yes"|] (fun _ -> ())
	in
	  if retval != 0 then
	    let msg =
	      Printf.sprintf "smart/deletecfg: smart channel failed: %i" retval
	    in
	    raise (Runtime_error (msg, output))
      with
	Job.Internal_error (msg, stdouterr) ->
	  let msg = 
	    Printf.sprintf "smart/deletecfg: Job.Internal_error: %s" msg
	  in
	  raise (Runtime_error (msg, stdouterr))
      | Job.Died_unexpected (msg, stdouterr)->
	  let msg =
	    Printf.sprintf "smart/deletecfg: Job.Died_unexpected: %s" msg
	  in
	  raise (Runtime_error (msg, stdouterr))
    and add_configuration name url =
      Debug.print 6 ("pkgsyssmart: add_configuration called (name=" ^
			name ^ ", url=" ^ url);
      try
	let retval, output =
	  Job.launch "smart" [|"aus:smart"; "channel"; "--yes"; "--add";
			       name; "type=rpm-md"; "baseurl=" ^ url|]
	    (fun _ -> ())
	in
	  if retval != 0 then
	    let msg =
	      Printf.sprintf "smart/addcfg: smart channel failed: %i" retval
	    in
	    raise (Runtime_error (msg, output))
      with
	Job.Internal_error (msg, stdouterr) ->
	  let msg = 
	    Printf.sprintf "smart/addcfg: Job.Internal_error: %s" msg
	  in
	  raise (Runtime_error (msg, stdouterr))
      | Job.Died_unexpected (msg, stdouterr)->
	  let msg =
	    Printf.sprintf "smart/addcfg: Job.Died_unexpected: %s" msg
	  in
	  raise (Runtime_error (msg, stdouterr))
    and name_of_url url =
      let pathonly =
	Str.replace_first (Str.regexp "^[^:/]+://[^/]+/") "" url
      in let pathlist = Str.split (Str.regexp "[/]") pathonly
      in
	   String.concat "_" (List.rev pathlist)
    in
      delete_configuration ();
      List.iter (fun url -> add_configuration (name_of_url url) url) repo_urls

  method update_metadata ?(progress_callback=(fun _ _ -> ())) () =
    Debug.print 4 "pkgsyssmart: update called";
    try
      let retval, output =
	Job.launch "smart" [|"aus:smart"; "update"|]
	  (trace_update_callback progress_callback)
      in
	if retval != 0 then
	  let msg =
	    Printf.sprintf "smart/install: smart update failed: %i" retval
	  in
	  raise (Runtime_error (msg,  output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "smart/update_metadata: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "smart/update_metadata: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method upgrade ?(progress_callback=(fun _ _ -> ())) () =
    Debug.print 4 "pkgsyssmart: upgrade called";
    try
      let retval, output =
	Job.launch "smart" [| "aus:smart"; "upgrade"; "--yes" |]
	  (trace_transaction_callback progress_callback)
      in
	if retval != 0 then
	  let msg = 
	    Printf.sprintf "smart/install: smart upgrade failed %i" retval
	  in
	  raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "smart/upgrade: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "smart/upgrade: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method install ?(progress_callback=(fun _ _ -> ())) pkgs =
    try
      Debug.print 4 "pkgsyssmart: install called";
      let retval, output =
	Job.launch "smart" (Array.of_list ("aus:smart" :: "install" :: "--yes"
					    :: List.map (string_of_name_arch
							    ~sep:"@") pkgs))
	  (trace_transaction_callback progress_callback)
      in
	if retval != 0 then
	  let msg = 
	    Printf.sprintf "smart/install: smart install failed: %i" retval
	  in
	  raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "smart/install: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "smart/install: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))


  method remove pkgs =
    raise (Runtime_error ("smart/remove: Not yet implemented", []))

  method clean () =
    try
      let retval, output =
	Job.launch "smart" [|"aus:smart"; "clean"|] (fun _ -> ())
      in
	if retval != 0 then
	  let msg = 
	    Printf.sprintf "smart/clean: smart clean packages failed: %i" retval
	  in
	  raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "smart/clean: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "smart/clean: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method distclean () =
    try
      let retval, output =
	Job.launch "smart" [|"aus:smart"; "clean"|] (fun _ -> ())
      in
	if retval != 0 then
	  let msg =
	    Printf.sprintf "smart/clean: smart clean all failed: %i" retval
	  in
	  raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "smart/distclean: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "smart/distclean: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
end
