(* $Id: pkgsysyum.ml,v 1.7 2016/01/08 14:25:38 heinz Exp $
 *
 * AUS - the advanced update system
 * Copyright (C) 2004-2005  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Printf

open Pkgsys
open Util

(* neede previous to ocaml 3.10 *)
let is_directory str =
  (Unix.stat str).Unix.st_kind = Unix.S_DIR

let whitespace_regexp = Str.regexp "[ \t]+"

let string_of_strings strs =
  List.fold_right (fun astr strs -> astr ^ "," ^ strs) strs ""

(** FIXME make this a config option *)
let yum_config_location = "/etc/yum.repos.d"

let trace_callback progress_callback line =
  let rx_start = Str.regexp "^Setting up .+ Process"
  and rx_download = Str.regexp "^[(] *\\([0-9]+\\)/ *\\([0-9]+\\)[)]:"
  and rx_install = Str.regexp "^[[] *\\([0-9]+\\)/ *\\([0-9]+\\)]:"
  and extract_numbers line = (* string_match must be called before! *)
    (float_of_string (Str.matched_group 1 line)) /.
      (float_of_string (Str.matched_group 2 line))
  in
    Debug.print 99 ("YUMJOB: " ^ line);
    if Str.string_match rx_start line 0 then
      progress_callback 0.0 "initializing.."
    else if Str.string_match rx_download line 0 then
      progress_callback (extract_numbers line) "(downloading)"
    else if Str.string_match rx_install line 0 then
      progress_callback (extract_numbers line) "(installing)"

(* hack, fixme: this is a dup and ugly, too *)
type loc = Loc of int * string
exception Parse_error of loc * string
let parse_int ?(range=fun _ -> true) str opname =
  try
      let i = int_of_string str
      in
        if (range i) = false then
          raise (Parse_error (Loc (0, "unknown"), sprintf
                                "options %s argument %s is out of range"
                                opname str));
        i
    with
        _ -> raise (Parse_error (Loc (0, "unknown"), sprintf
                                   "option %s expects int but got %s"
                                   opname str))

class pkgSysYum = object
  inherit packageSystem

  val mutable repo_urls = []

  method init () = repo_urls <- []

(* even the debug pkgsys must honor the repository location to enable
   parsing of default group files
*)
  method additional_includes () =
    List.rev (List.map (fun x -> (fst x) ^ "/ausgroups.conf") repo_urls)
  method configure msgs =
    let cfg_msg msg =
      match Str.split whitespace_regexp msg with
	| "repository" :: url :: [] -> repo_urls <- (url, 99) :: repo_urls
	| "repository" as n :: url :: pri :: [] -> repo_urls <- (url, parse_int ~range:(fun x -> x >= 0 && x <= 99) pri n) :: repo_urls
	| _ -> ()
    in
      List.iter cfg_msg msgs

  (* move away all "evil" config files and install our owns *)
  method activate_configuration () =
    let configfiles = Sys.readdir yum_config_location
    in let handle_configfile path cf =
      try
	if cf != "aus.repo" then begin (* will be overwritten anyway *)
	  if not (is_directory (Filename.concat path cf)) then
	    Sys.remove (Filename.concat path cf)
	end
      with
	  x ->
	    raise (Runtime_error (sprintf "%s go unknown exception: %s"
				     "yum/activate_configuration"
				     (Printexc.to_string x), []))
    in
      Array.iter (handle_configfile yum_config_location) configfiles;
      let ofi = open_out (Filename.concat yum_config_location "aus.repo")
      in let create_repo (url, pri) =
	let name_of_url url =
	  let pathonly =
	    Str.replace_first (Str.regexp "^[^:/]+://[^/]+/") "" url
	  in let pathlist = Str.split (Str.regexp "[/]") pathonly
	  in
	       String.concat "_" (List.rev pathlist)
	in
	  fprintf ofi "[%s]\nname=%s\nbaseurl=%s\nenabled=1\ngpgcheck=0\npriority=%i\n\n"
	    (name_of_url url) ("AUS Repository " ^ (name_of_url url)) url pri
      in
	fprintf ofi "# %s\n# %s\n\n\n"
	  "This file is automatically created and updated by AUS."
	  "Manual modifications will be ignored and overwritten.";
	List.iter create_repo repo_urls;
	close_out ofi

  method update_metadata ?(progress_callback=(fun _ _ -> ())) () =
    try
      let retval, output =
	Job.launch "yum" [|"aus:yum"; "makecache"; "fast"|] (fun _ -> ())
      in
      if retval != 0 then
	let msg = 
	  Printf.sprintf "yum/update_metadata: yum makecache failed: %i" retval
	in
	raise (Runtime_error (msg, output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "yum/update_metadata: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "yum/update_metadata: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method upgrade ?(progress_callback=(fun _ _ -> ())) () =
    Debug.print 4 "pkgsysyum: upgrade called";
    try
      let retval, output =
	Job.launch "yum" [|"aus:yum"; "-y"; "-d"; "10"; "--skip-broken"; "upgrade"|]
	  (fun _ -> ())
(*	  (trace_callback progress_callback) *)
      in
	if retval != 0 then
	  raise (Runtime_error
		    ("yum/install: yum upgrade failed",  output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "yum/upgrade: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "yum/upgrade: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method install ?(progress_callback=(fun _ _ -> ())) pkgs =
    try
      Debug.print 4 "pkgsysyum: install called";
      let retval, output =
	Job.launch "yum" (Array.of_list ("aus:yum" :: "-y" :: "-d" :: "10" :: "--skip-broken" :: "install" ::
					    List.map (string_of_name_arch
							 ~sep:".") pkgs))
	  (fun _ -> ())
(*	  (trace_callback progress_callback) *)
      in
	if retval != 0 then
	  raise (Runtime_error
		    ("yum/install: yum install failed",  output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "yum/install: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "yum/install: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method remove pkgs =
    raise (Runtime_error ("yum/remove: Not yet implemented", []))

  method clean () =
    try
      let retval, output =
	Job.launch "yum" [|"aus:yum"; "clean"; "expire-cache"; "packages"|] (fun _ -> ())
      in
	if retval != 0 then
	  raise (Runtime_error ("yum/clean: yum clean expire-cache packages failed",
			       output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "yum/clean: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "yum/clean: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))

  method distclean () =
    try
      let retval, output =
	Job.launch "yum" [|"aus:yum"; "clean"; "all"|] (fun _ -> ())
      in
	if retval != 0 then
	  raise (Runtime_error ("yum/clean: yum clean all failed",
			       output))
    with
      Job.Internal_error (msg, stdouterr) ->
	let msg = 
	  Printf.sprintf "yum/distclean: Job.Internal_error: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
    | Job.Died_unexpected (msg, stdouterr)->
	let msg =
	  Printf.sprintf "yum/distclean: Job.Died_unexpected: %s" msg
	in
	raise (Runtime_error (msg, stdouterr))
end
