(* $Id: rfile.ml,v 1.9 2016/01/08 14:25:38 heinz Exp $
 *
 * AUS - the advanced update system
 * Copyright (C) 2004  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

type t = { cache_path: string; cache: (string, bool) Hashtbl.t }

exception Illegal_url of string
  (** *)

exception Connection_failure of string
  (** URL may be valid, but connection could not be made (e.g. network failure)
  *)

exception Url_not_found of string
  (** Raised if URL is not found (e.g. 404) *)

exception Cache_failure of string
  (** Error accessing cache. Fatal internal error, should not happen *)


let wget404_regexp =
  Str.regexp "\\(No such file\\)\\|\\(ERROR 404: Not Found\\)\\|\\(failed\\)"

let url_map str =
  let str = Str.global_replace (Str.regexp "%") "%%" str
  in let str = Str.global_replace (Str.regexp "/") "%I" str
  in
    Str.global_replace (Str.regexp "~") "%-" str

let make path =
  { cache_path = path; cache = Hashtbl.create 8; }

let wget_update_cache t url =
  let wget_call =
    Printf.sprintf "wget -nv -N -O %s/%s %s" t.cache_path (url_map url) url
  in
    Debug.print 2 (Printf.sprintf "Rfile.wget_update_cache: %s" wget_call);
    let (i,o,e) = Unix.open_process_full wget_call [||]
    in let rec test404 ich =
	try
	  if Str.string_match wget404_regexp (input_line ich) 0 then begin
	    ignore (Unix.close_process_full (i, o, e));
	    raise (Url_not_found url)
	  end else
	    test404 ich
	with
	    End_of_file ->
	      match Unix.close_process_full (i, o, e) with
		  Unix.WEXITED 0 -> url_map url
		| _ -> raise (Connection_failure url)
    in
      test404 e

let open_in_channel t url =
  let rec open_file path =
    try
      open_in path
    with
	Sys_error _ -> raise (Url_not_found url)
  and open_cached name =
    try
      open_in (t.cache_path ^ "/" ^ name)
    with
	Sys_error _ -> raise (Cache_failure url)
  and open_wget url =
    try
      if Hashtbl.find t.cache (url_map url) then
	open_cached (url_map url)
      else
	raise (Url_not_found url)
    with
	Not_found ->
	  open_cached (wget_update_cache t url)
  in
    if url = "" then
      raise (Url_not_found url);
    if (url.[0] = '/') || (url.[0] = '.') then
      open_file url
    else match Str.bounded_split (Str.regexp "://") url 2 with
	[ "file"; path ] -> open_file ("/" ^ path)
      | [ "http"; _ ]
      | [ "ftp"; _ ] -> open_wget url
      | _ -> raise (Illegal_url url)

